/*
 * lcdc.c
 *
 *  Created on: 2013/04/24
 *  Copyright(C) 2013 Ryoichi Yokomizu
 *  All rights reserved
 */

#include <string.h>
#include "lcdc.h"
#include "spi.h"

/* ハードウェアのコンフィグレーション */
#define BUS_8BITALIGN	/* 8bit転送が可能な場合に定義 */
#define HVLINE_OPTIMIZE	/* 水平線・垂直線描画の高速化 (パラレルバス接続の場合は要評価) */

/* S1D13781レジスタ定義 */
#define LCDC_REG_FRAMEBUF	0x00000000
#define LCDC_REG_LUT1		0x00060000
#define LCDC_REG_LUT2		0x00060400
#define LCDC_REG_REGISTER	0x00060800

#define LCDC_SCREEN_WIDTH	480
#define LCDC_SCREEN_HEIGHT	272
#define LCDC_DEFAULT_BITS	24

#define LCDC_FONT_WIDTH		5
#define LCDC_FONT_HEIGHT	8
#define LCDC_FONT_MAXMAG	16

/* 内部関数 */
static void lcdc_plot4points(short cx, short cy, short x, short y, unsigned char r, unsigned char g, unsigned char b);

/* 内部変数 */
static unsigned char g_mode = LCDC_MODE_UNKNOWN;
static unsigned long g_mainadr;
static unsigned short g_bytesperpixel;

/* 外部変数の参照 */
extern const unsigned char g_FontData5x8[][8];

/* LCDCと内蔵周辺デバイスの初期化 */
void LCDC_Init(unsigned char mode)
{
	switch (mode) {
	case LCDC_MODE_RGB888:
		g_bytesperpixel = (unsigned short)3;
		break;
	case LCDC_MODE_RGB565:
		g_bytesperpixel = (unsigned short)2;
		break;
	case LCDC_MODE_LUT8:
		g_bytesperpixel = (unsigned short)1;
		mode = LCDC_MODE_UNKNOWN;	/* 未実装 */
		break;
	default:
		mode = LCDC_MODE_UNKNOWN;
		g_bytesperpixel = (unsigned short)LCDC_DEFAULT_BITS / 8;
		break;
	}

	spi_init(LCDC_SCREEN_WIDTH * LCDC_SCREEN_HEIGHT * g_bytesperpixel);

	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x12, 0x000f);		/* MM=16 */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x14, 0x0029);		/* LL=42 */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x10, 0x0001);		/* PLL enable */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x16, 0x0006);		/* fPLL_REF_CLK=fCLKI/7 */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x20, 0x004f);		/* DE: Hi-active,
																   PCLK Polarity: Rising edge,
																   PanelType: Color TFT 24-bit */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x24, 0x003c);		/* Horizontal Display Width Register : 480 */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x26, 0x002d);		/* Horizontal Non-Display Period Register: 45 */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x28, 0x0110);		/* Vertical Display Height Register: 272 */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x2A, 0x0010);		/* Vertical Non-Display Period Register: 16 */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x04, 0x0002);		/* MCLK/PCLK enable */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x22, 0x0021);		/* Panel interface enable, TE=Enable */

	switch (mode) {
	case LCDC_MODE_RGB888:
		LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x40, 0x0000);	/* RGB8:8:8 */
		break;
	case LCDC_MODE_RGB565:
		LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x40, 0x0001);	/* RGB5:6:5 */
		break;
	case LCDC_MODE_LUT8:
	default:
		break;
	}
	g_mode = mode;
	g_mainadr = (unsigned long)0;

	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x42, g_mainadr & 0x0000ffff);			/* Main Layer start address(L) */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x44, (g_mainadr >> 16) & 0x0007);		/* Main Layer start address(H) */

	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x80, 0x0080);	/* 2D BitBltエンジンリセット */

	LCDC_ClearScreen();
}

/* 表示クリア */
void LCDC_ClearScreen(void)
{
	switch (g_mode) {
	case LCDC_MODE_RGB888:
		LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x82, 0x000b);	/* 24bpp, リニアブロック */
		LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x9a, 0x0000);	/* 前景色:Blue,Green=0 */
		LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x9c, 0x0000);	/*        Red = 0 */
		break;
	case LCDC_MODE_RGB565:
		LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x82, 0x0007);	/* 16bpp, リニアブロック */
		LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x9a, 0x0000);	/* 前景色:Red,Green,Blue=0 */
		break;
	case LCDC_MODE_LUT8:
	default:
		return;
	}
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x86, 0x0002);						/* コマンド: ソリッドフィル */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x8c, g_mainadr & 0x0000ffff);		/* 転送先スタートアドレス */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x8e, (g_mainadr >> 16) & 0x0007);
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x92, LCDC_SCREEN_WIDTH);			/* Width :480 */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x94, LCDC_SCREEN_HEIGHT);			/* Height:272 */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x80, 0x0001);						/* Start */

	while ((LCDC_ReadRegister16(LCDC_REG_REGISTER + 0x84) & 0x0001) != 0);	/* 完了待ち */
}

/* 描画関数 */
void LCDC_BoxFill(short x1, short y1, short x2, short y2, unsigned char r, unsigned char g, unsigned char b)
{
	unsigned long startadr;
	short temp;

	/* 前回の操作を完了待ち */
	while ((LCDC_ReadRegister16(LCDC_REG_REGISTER + 0x84) & 0x0001) != 0);	/* 完了待ち */

	/* x1, x2座標の正規化 */
	if (x1 > x2) {
		temp = x1;
		x1 = x2; x2 = temp;
	}

	/* y1, y2座標の正規化 */
	if (y1 > y2) {
		temp = y1;
		y1 = y2; y2 = temp;
	}

	switch (g_mode) {
	case LCDC_MODE_RGB888:
		LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x82, 0x0008);	/* 24bpp */
		LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x9a, (unsigned short)b | (unsigned short)(g << 8));	/* 前景色:Blue,Green */
		LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x9c, (unsigned short)r);								/*       Red */
		startadr = g_mainadr + (y1 * LCDC_SCREEN_WIDTH + x1) * 3;
		break;
	case LCDC_MODE_RGB565:
		b >>= 3;
		g >>= 2;
		r >>= 3;
		LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x82, 0x0004);	/* 16bpp */
		LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x9a, (r << 11) | (g << 5) | b);	/* 前景色:Red,Green,Blue */
////////LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x9c, (unsigned short)0);								/*       Red */
		startadr = g_mainadr + (y1 * LCDC_SCREEN_WIDTH + x1) * 2;
		break;
	case LCDC_MODE_LUT8:
	default:
		return;
	}
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x86, 0x0002);		/* コマンド: ソリッドフィル */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x8c, (unsigned short)(startadr & 0xffff));			/* 転送先スタートアドレス */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x8e, (unsigned short)((startadr >> 16) & 0x0007));	/* 転送先スタートアドレス */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x90, (unsigned short)LCDC_SCREEN_WIDTH);	// 矩形オフセットを設定
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x92, x2 - x1 + 1);	/* Width */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x94, y2 - y1 + 1);	/* Height */

	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x80, 0x0001);		/* Start */
}

void LCDC_BitBlt(short x, short y, short width, short height, short srcx, short srcy)
{
	unsigned long src_startadr, dst_startadr;
	unsigned short cmd;

	if ((x < 0) || (y < 0) || (width <= 0) || (height <= 0)) {
		return;
	}

	/* 前回の操作を完了待ち */
	while ((LCDC_ReadRegister16(LCDC_REG_REGISTER + 0x84) & 0x0001) != 0);	/* 完了待ち */

	switch (g_mode) {
	case LCDC_MODE_RGB888:
		LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x82, 0x0008);	/* 24bpp */
		break;
	case LCDC_MODE_RGB565:
		LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x82, 0x0004);	/* 16bpp */
		break;
	case LCDC_MODE_LUT8:
	default:
		return;
	}
	
	src_startadr = g_mainadr + (srcy * LCDC_SCREEN_WIDTH + srcx) * g_bytesperpixel;
	dst_startadr = g_mainadr + (y * LCDC_SCREEN_WIDTH + x) * g_bytesperpixel;

	if (src_startadr < dst_startadr) {
		cmd = 0x0001;	/* ムーブネガティブ */
		src_startadr = g_mainadr + ((srcy + height - 1) * LCDC_SCREEN_WIDTH + srcx + width - 1) * g_bytesperpixel + (g_bytesperpixel - 1);
		dst_startadr = g_mainadr + ((y + height - 1) * LCDC_SCREEN_WIDTH + x + width - 1) * g_bytesperpixel + (g_bytesperpixel - 1);
	}
	else {
		cmd = 0x0000;	/* ムーブポジティブ */
	}

	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x86, cmd);		/* コマンド */

	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x88, (unsigned short)(src_startadr & 0xffff));			/* 転送元スタートアドレス */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x8a, (unsigned short)((src_startadr >> 16) & 0x0007));	/* 転送元スタートアドレス */

	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x8c, (unsigned short)(dst_startadr & 0xffff));			/* 転送先スタートアドレス */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x8e, (unsigned short)((dst_startadr >> 16) & 0x0007));	/* 転送先スタートアドレス */
	
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x90, (unsigned short)LCDC_SCREEN_WIDTH);	/* 矩形オフセット */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x92, width);	/* Width */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x94, height);	/* Height */

	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x80, 0x0001);		/* Start */
}

void LCDC_Rectangle(short x1, short y1, short x2, short y2, unsigned char r, unsigned char g, unsigned char b)
{
	unsigned short temp;

	/* x1, x2座標の正規化 */
	if (x1 > x2) {
		temp = x1;
		x1 = x2; x2 = temp;
	}

	/* y1, y2座標の正規化 */
	if (y1 > y2) {
		temp = y1;
		y1 = y2; y2 = temp;
	}

	LCDC_HLine(x1, y1, x2, r, g, b);
	LCDC_HLine(x1, y2, x2, r, g, b);
	LCDC_VLine(x1, y1, y2, r, g, b);
	LCDC_VLine(x2, y1, y2, r, g, b);
}

void LCDC_HLine(short x1, short y1, short x2, unsigned char r, unsigned char g, unsigned char b)
{
#ifdef HVLINE_OPTIMIZE
	/* ハードウェア描画 */
	if (x1 == x2) {
		LCDC_SetPixel(x1, y1, r, g, b);
	}
	else {
		LCDC_BoxFill(x1, y1, x2, y1, r, g, b);
	}
#else
	/* ソフトウェア描画 */
	unsigned short count;

	for (count = 0; count < x2 - x1 + 1; count++) {
		LCDC_SetPixel(x1 + count, y1, r, g, b);
	}
#endif
}

void LCDC_VLine(short x1, short y1, short y2, unsigned char r, unsigned char g, unsigned char b)
{
#ifdef HVLINE_OPTIMIZE
	/* ハードウェア描画 */
	if (y1 == y2) {
		LCDC_SetPixel(x1, y1, r, g, b);
	}
	else {
		LCDC_BoxFill(x1, y1, x1, y2, r, g, b);
	}
#else
	/* ソフトウェア描画 */
	unsigned short count;

	for (count = 0; count < y2 - y1 + 1; count++) {
		LCDC_SetPixel(x1, y1 + count, r, g, b);
	}
#endif
}

#define DIFF(a, b) ((a) < (b) ? (b) - (a): (a) - (b))

void LCDC_Line(short x1, short y1, short x2, short y2, unsigned char r, unsigned char g, unsigned char b)
{
	short dx, dy, sx, sy, err, e2;

	dx = DIFF(x1, x2);
	dy = DIFF(y1, y2);

	if (x1 < x2) {
		sx = 1;
	}
	else {
		sx = -1;
	}

	if (y1 < y2) {
		sy = 1;
	}
	else {
		sy = -1;
	}

	err = dx - dy;

	while(1) {
		LCDC_SetPixel(x1, y1, r, g, b);
		if ((x1 == x2) && (y1 == y2)) {
			break;
		}
		e2 = 2 * err;
		if (e2 > -dy) {
			err = err - dy;
			x1 += sx;
		}
		if (e2 < dx) {
			err += dx;
			y1 += sy;
		}
	}
}

void LCDC_Circle(short cx, short cy, short rn, unsigned char r, unsigned char g, unsigned char b)
{
	short error = -rn;
	short x = rn;
	short y = 0;

	while (x >= y) {

		lcdc_plot4points(cx, cy, x, y, r, g, b);
		if (x != y) {
			lcdc_plot4points(cx, cy, y, x, r, g, b);
		}

		error += y;
		++y;
		error += y;

		if (error >= 0) {
			error -= x;
			--x;
			error -= x;
		}
	}
}

static void lcdc_plot4points(short cx, short cy, short x, short y, unsigned char r, unsigned char g, unsigned char b)
{
	LCDC_SetPixel(cx + x, cy + y, r, g, b);
	if (x != 0) {
		LCDC_SetPixel(cx - x, cy + y, r, g, b);
	}
	if (y != 0) {
		LCDC_SetPixel(cx + x, cy - y, r, g, b);
	}
	if ((x != 0) && (y != 0)) {
		LCDC_SetPixel(cx - x, cy - y, r, g, b);
	}
}

void LCDC_CircleFill(short cx, short cy, short rn, unsigned char r, unsigned char g, unsigned char b)
{
	short error = -rn;
	short x = rn;
	short y = 0;

	while (x >= y) {

		LCDC_HLine(cx - x, cy - y, cx + x, r, g, b);
		LCDC_HLine(cx - x, cy + y, cx + x, r, g, b);
		LCDC_HLine(cx - y, cy - x, cx + y, r, g, b);
		LCDC_HLine(cx - y, cy + x, cx + y, r, g, b);

		error += y;
		++y;
		error += y;

		if (error >= 0) {
			error -= x;
			--x;
			error -= x;
		}
	}
}

void LCDC_SetPixel(short x1, short y1, unsigned char r, unsigned char g, unsigned char b)
{
	unsigned long addr;
	unsigned short val1, val2, temp;
#ifdef BUS_8BITALIGN
	unsigned long data24;
#endif

	if ((x1 < 0) || (x1 >= LCDC_SCREEN_WIDTH)) {
		return;
	}
	if ((y1 < 0) || (y1 >= LCDC_SCREEN_HEIGHT)) {
		return;
	}

	switch (g_mode) {
	case LCDC_MODE_RGB888:
		addr = g_mainadr + (y1 * LCDC_SCREEN_WIDTH + x1) * 3;
#ifdef BUS_8BITALIGN
		/* 8bit転送が可能な場合 */
		data24 = (unsigned long)r | ((unsigned long)g << 8) | ((unsigned long)b << 16);
		LCDC_WriteRegister24(addr, data24);
#else
		/* 16bitバス */
		if ((addr % 2) == 0) {
			val1 = (unsigned short)b | ((unsigned short)g << 8);
			val2 = (unsigned short)r;
			LCDC_WriteRegister16(addr, val1);
			temp = LCDC_ReadRegister16(addr + 2);
			temp = (temp & 0xff00) | val2;	/* 次のピクセルのB要素を保護する */
			LCDC_WriteRegister16(addr + 2, temp);
		}
		else {
			addr &= 0xfffffffe;
			val1 = (unsigned short)b << 8;
			val2 = ((unsigned short)r << 8) | (unsigned short)g;
			temp = LCDC_ReadRegister16(addr);
			temp = (temp & 0x00ff) | val1;
			LCDC_WriteRegister16(addr, temp);
			LCDC_WriteRegister16(addr + 2, val2);
		}
#endif
		break;
	case LCDC_MODE_RGB565:
		addr = g_mainadr + (y1 * LCDC_SCREEN_WIDTH + x1) * 2;
		b >>= 3;
		g >>= 2;
		r >>= 3;
		LCDC_WriteRegister16(addr, (unsigned short)((r << 11) | (g << 5) | b));
		break;
	case LCDC_MODE_LUT8:
		break;
	}
}

void LCDC_GetPixel(short x1, short y1, unsigned char *pr, unsigned char *pg, unsigned char *pb)
{
	unsigned long addr;
	unsigned short temp;
#ifdef BUS_8BITALIGN
	unsigned long data24;
#endif

	switch (g_mode) {
	case LCDC_MODE_RGB888:
		addr = g_mainadr + (y1 * LCDC_SCREEN_WIDTH + x1) * 3;
#ifdef BUS_8BITALIGN
		data24 = LCDC_ReadRegister24(addr);
		*pr = data24 & 0x000000ff;
		*pg = (data24 & 0x0000ff00) >> 8;
		*pb = (data24 & 0x00ff0000) >> 16;
#else
		if (addr % 2 == 0) {
			temp = LCDC_ReadRegister16(addr);
			*pb = temp & 0x00ff;
			*pg = (temp & 0xff00) >> 8;
			temp = LCDC_ReadRegister16(addr);
			*pr = temp & 0x00ff;
		}
		else {
			addr &= 0xfffffffe;
			temp = LCDC_ReadRegister16(addr);
			*pb = (temp & 0xff00) >> 8;
			temp = LCDC_ReadRegister16(addr);
			*pg = temp & 0x00ff;
			*pr = (temp & 0xff00) >> 8;
		}
#endif
		break;
	case LCDC_MODE_RGB565:
		addr = g_mainadr + (y1 * LCDC_SCREEN_WIDTH + x1) * 2;
		temp = LCDC_ReadRegister16(addr);
		*pr = (temp & 0xf800) >> 8;
		*pg = (temp & 0x07e0) >> 3;
		*pb = (temp & 0x001f) << 3;
		break;
	case LCDC_MODE_LUT8:
		break;
	}
}

void LCDC_DrawFont(short x1, short y1, short xmag, short ymag, unsigned char code, unsigned char r, unsigned char g, unsigned char b)
{
	unsigned long addr;
	unsigned short val1, val2, temp, draw_width, draw_height;
	short x, y;
	unsigned char *pfnt;
	unsigned char bittbl[LCDC_FONT_WIDTH * LCDC_FONT_MAXMAG];
	unsigned char bittbl_src[8] = {
		0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01
	};
#ifdef BUS_8BITALIGN
	unsigned long data24;
#endif

	if ((x1 < 0) || (x1 >= LCDC_SCREEN_WIDTH)) {
		return;
	}
	if ((y1 < 0) || (y1 >= LCDC_SCREEN_HEIGHT)) {
		return;
	}

	if ((xmag < 1) || (xmag > LCDC_FONT_MAXMAG)) {
		return;
	}

	if ((ymag < 1) || (ymag > LCDC_FONT_MAXMAG)) {
		return;
	}

	if ((code < 0x20) || (code > 0x7e)) {
		return;
	}

	draw_width = LCDC_SCREEN_WIDTH - x1;
	if (draw_width > LCDC_FONT_WIDTH * xmag) {
		draw_width = LCDC_FONT_WIDTH * xmag;
	}

	draw_height = LCDC_SCREEN_HEIGHT - y1;
	if (draw_height > LCDC_FONT_HEIGHT * ymag) {
		draw_height = LCDC_FONT_HEIGHT * ymag;
	}

	for (x = 0; x < draw_width; x++) {
		bittbl[x] = bittbl_src[x / xmag];
	}

	switch (g_mode) {
	case LCDC_MODE_RGB888:
		for (y = 0; y < draw_height; y++) {
			pfnt = (unsigned char *)&g_FontData5x8[code - 0x20][y / ymag];
			addr = g_mainadr + ((y1 + y) * LCDC_SCREEN_WIDTH + x1) * 3;
			for (x = 0; x < draw_width; x++) {
				if ((*pfnt & bittbl[x]) != 0) {
#ifdef BUS_8BITALIGN
					/* 8bit転送が可能な場合 */
					data24 = (unsigned long)r | ((unsigned long)g << 8) | ((unsigned long)b << 16);
					LCDC_WriteRegister24(addr, data24);
#else
					/* 16bitバス */
					if ((addr % 2) == 0) {
						val1 = (unsigned short)b | ((unsigned short)g << 8);
						val2 = (unsigned short)r;
						LCDC_WriteRegister16(addr, val1);
						temp = LCDC_ReadRegister16(addr + 2);
						temp = (temp & 0xff00) | val2;	/* 次のピクセルのB要素を保護する */
						LCDC_WriteRegister16(addr + 2, temp);
					}
					else {
						val1 = (unsigned short)b << 8;
						val2 = ((unsigned short)r << 8) | (unsigned short)g;
						temp = LCDC_ReadRegister16(addr - 1);
						temp = (temp & 0x00ff) | val1;
						LCDC_WriteRegister16(addr - 1, temp);
						LCDC_WriteRegister16(addr + 1, val2);
					}
#endif
				}
				addr += 3;
			}
		}
		break;
	case LCDC_MODE_RGB565:
		b >>= 3; g >>= 2; r >>= 3;
		val1 = (unsigned short)((r << 11) | (g << 5) | b);
		for (y = 0; y < draw_height; y++) {
			pfnt = (unsigned char *)&g_FontData5x8[code - 0x20][y / ymag];
			addr = g_mainadr + ((y1 + y) * LCDC_SCREEN_WIDTH + x1) * 2;
			for (x = 0; x < draw_width; x++) {
				if ((*pfnt & bittbl[x]) != 0) {
					LCDC_WriteRegister16(addr, val1);
				}
				addr += 2;
			}
		}
		break;
	case LCDC_MODE_LUT8:
		break;
	}
}

void LCDC_DrawString(short x1, short y1, short xmag, short ymag, char *pstring, unsigned char r, unsigned char g, unsigned char b)
{
	int count;

	for (count = 0; count < strlen((const char *)pstring); count++) {
		LCDC_DrawFont(x1, y1, xmag, ymag, *(pstring + count), r, g, b);
		x1 += (LCDC_FONT_WIDTH + 1) * xmag;
	}
}

void LCDC_Scroll(short h, short v)
{
	short x, y, width, height, srcx, srcy;
	if ((h == 0) && (v == 0)) {
		return;
	}
	
	if (v > 0) {
		y = 0;
		height = LCDC_SCREEN_HEIGHT - v;
		srcy = v;
	}
	else {
		y = -v;
		height = LCDC_SCREEN_HEIGHT + v;
		srcy = 0;
	}
	
	if (h > 0) {
		x = 0;
		width = LCDC_SCREEN_WIDTH - h;
		srcx = h;
	}
	else {
		x = -h;
		width = LCDC_SCREEN_WIDTH + h;
		srcx = 0;
	}
	
	LCDC_BitBlt(x, y, width, height, srcx, srcy);
}

/* VDISP期間待ち(チラ付き防止用) */
void LCDC_WaitVdisp(void)
{
	while ((LCDC_ReadRegister16(LCDC_REG_REGISTER + 0x22) & 0x0080) != 0);	/* 垂直非表示期間中(TE=1)は待つ */
	while ((LCDC_ReadRegister16(LCDC_REG_REGISTER + 0x22) & 0x0080) == 0);	/* 垂直表示期間中(TE=0)は待つ */
}

void LCDC_SetFrameStartAddress(unsigned long addr)
{
	g_mainadr = addr;

	/* 注意：次のフレームから有効 */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x42, g_mainadr & 0x0000ffff);			/* Main Layer start address(L) */
	LCDC_WriteRegister16(LCDC_REG_REGISTER + 0x44, (g_mainadr >> 16) & 0x0007);	/* Main Layer start address(H) */
}

unsigned long LCDC_GetFrameStartAddress(void)
{
	return g_mainadr;
}

/* 表示データ/LUT転送 */
/* void LCDC_BurstWrite(unsigned long addr, unsigned char* buf, unsigned long size) */
/* ===> spi.cへ移動 */

/* LCDCレジスタ書き込み */
/* void LCDC_WriteRegister16(unsigned long addr, unsigned short data) */
/* ===> spi.cへ移動 */

/* LCDCレジスタ読み込み */
/* unsigned short LCDC_ReadRegister16(unsigned long addr) */
/* ===> spi.cへ移動 */

/* LCDCの情報を取得 */
unsigned short LCDC_GetBitsPerPixel(void)
{
	unsigned short result = LCDC_DEFAULT_BITS;

	switch (g_mode) {
	case LCDC_MODE_RGB888:
		result = 24;
		break;
	case LCDC_MODE_RGB565:
		result = 16;
		break;
	case LCDC_MODE_LUT8:
		result = 8;
		break;
	default:
		break;
	}

	return result;
}

unsigned short LCDC_GetScreenWidth(void)
{
	return (unsigned short)LCDC_SCREEN_WIDTH;
}

unsigned short LCDC_GetScreenHeight(void)
{
	return (unsigned short)LCDC_SCREEN_HEIGHT;
}

unsigned short LCDC_GetImageFormat(void)
{
	return (unsigned short)g_mode;
}
